/**
  *  \file client/dlg-build.cc
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-build.h"
#include "client/winfo.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/spec.h"
#include "game/struct.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/table.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/msgbox.h"

/*************************** WComponentListbox ***************************/

/** \class WComponentListbox
    \brief List box containing list of ship parts

    This class only provides a method drawComponent() to draw ship
    parts. Derived classes use this to achieve a uniform look across
    all component lists. */

/** Contructor.
    \param id       widget id
    \param lines    number of lines of widget
    \param width    width of widget
    \param planet   planet. May be null. Ignored if it doesn't have a
                    starbase. Tech levels (and tick marks) are taken
                    from here. */
WComponentListbox::WComponentListbox(int id, int lines, int width, GPlanet* planet)
    : UIListbox(id, lines, width, 0), tick_planet(planet)
{
    if (!tick_planet->hasBase())
        tick_planet = 0;
}

/** Draw a component entry.
    \param can     target canvas
    \param line    logical line number (see drawEntry())
    \param com     component to draw.
    \param tech_area the tech area used by this component.
    \param have    amount of such components in storage (used for tick marks) */
void
WComponentListbox::drawComponent(GfxCanvas& can, int line, const GComponent& com, TTechLevel tech_area, int have)
{
    UIColor::Color c = (!tick_planet || com.getTech() <= tick_planet->getBaseTech(tech_area)
                        ? UIColor::tc_Static
                        : com.getTech() > 6 && !allowRegistered(tick_planet->getOwner())
                        ? UIColor::tc_Faded
                        : UIColor::tc_Blue);
    const char* tick = have ? "\7" : " ";
    drawEntry(can, c, line, tick + com.getName());
}

/****************************** WHullListbox *****************************/

/** \class WHullListbox
    \brief Hull List box

    This widget displays a list of hulls. You can (actually, must)
    provide a list of hulls to display. */

/** Constructor. \see WComponentListbox::WComponentListbox() */
WHullListbox::WHullListbox(int id, int lines, int width, GPlanet* p)
    : WComponentListbox(id, lines, width, p)
{ }

WHullListbox::~WHullListbox()
{ }

void
WHullListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        const int slot = i + getTop();
        if (slot < int(content.size())) {
            const int hull = content[slot];
            int have = 0;
            if (tick_planet && tick_planet->hasBase()) {
                int n = getTruehullSlot(tick_planet->getOwner(), hull);
                if (n)
                    have = tick_planet->getBaseHullStoreSlot(n);
            }
            drawComponent(can, i, getHull(hull), tl_Hull, have);
        } else {
            drawEntry(can, UIColor::tc_Static, i, string_t());
        }
    }
}

/** Add a hull to the end of the list.
    \param n hull number (1..NUM_HULLS)
    \see swapContents() */
void
WHullListbox::addHullId(int n)
{
    content.push_back(n);
    setSize(content.size());
    drawWidget();
}

/** Set content of this list.
    \param new_content vector of hull numbers (1..NUM_HULLS). After
    this function completes, this widget will display the old contents
    of /new_content/, and /new_content/ will contain the old content. */
void
WHullListbox::swapContents(std::vector<int>& new_content)
{
    int now = getCurrentHull();
    content.swap(new_content);
    setSize(content.size());
    scrollToHull(now);
    adjustPosition();
    drawWidget();
}

/** Get currently-selected hull number. */
int
WHullListbox::getCurrentHull() const
{
    if (getCurrent() < int(content.size()))
        return content[getCurrent()];
    else
        return 0;
}

/** Scroll to specified hull. Does nothing if this hull is not on our
    list. */
void
WHullListbox::scrollToHull(int id)
{
    for (std::vector<int>::size_type i = 0; i < content.size(); ++i) {
        if (id == content[i]) {
            scrollTo(i, 1);
            return;
        }
    }
}

/************ WEngineListbox / WLauncherListbox / WBeamListbox ***********/

WEngineListbox::WEngineListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_ENGINES, width, tick_planet)
{
    setSize(NUM_ENGINES);
}

void
WEngineListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        if (slot <= NUM_ENGINES)
            drawComponent(can, i, getEngine(slot), tl_Engine,
                          tick_planet ? tick_planet->getBaseEngineStore(slot) : 0);
        else
            drawEntry(can, UIColor::tc_Static, i, string_t());
    }
}

WBeamListbox::WBeamListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_BEAMS, width, tick_planet)
{
    setSize(NUM_BEAMS);
}

void
WBeamListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        if (slot <= NUM_BEAMS)
            drawComponent(can, i, getBeam(slot), tl_Beam,
                          tick_planet ? tick_planet->getBaseBeamStore(slot) : 0);
        else
            drawEntry(can, UIColor::tc_Static, i, string_t());
    }
}

WLauncherListbox::WLauncherListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_TORPS, width, tick_planet)
{
    setSize(NUM_TORPS);
}

void
WLauncherListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        if (slot <= NUM_TORPS)
            drawComponent(can, i, getTorp(slot), tl_Torpedo,
                          tick_planet ? tick_planet->getBaseLauncherStore(slot) : 0);
        else
            drawEntry(can, UIColor::tc_Static, i, string_t());
    }
}

/***************************** WOrderDisplay *****************************/

/** \class WOrderDisplay
    \brief Display a Build Order

    This will display a build order on the ship build screen. */
WOrderDisplay::WOrderDisplay()
{
    o.hull = 0;
}

void
WOrderDisplay::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);

    if (o.hull) {
        GHull& h = getHull(o.hull);

        tw.newLColumn();
        tw.writeVariableCell(UIColor::tc_Contrast, _("Your order:"));
        tw.writeVariableCell(UIColor::tc_Static, "\7 " + h.getName());
        tw.writeVariableCell(UIColor::tc_Static,
                             /* this format string turns into
                                "Transwarp Drive" or "7 x Transwarp Drive" */
                             format(_("\7 %!d%!1{%0$d x %}%1$s"),
                                    h.getNumEngines(),
                                    getEngine(o.engine).getName()));
        if (o.beam_count)
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\7 %!d%!1{%0$d x %}%1$s"),
                                        o.beam_count,
                                        getBeam(o.beam_type).getName()));
        if (o.torp_count)
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\7 %!d%!1{%0$d x %}%1$s"),
                                        o.torp_count,
                                        getTorp(o.torp_type).getName()));
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\7 %d fighter bay%!1{s%}"), h.getNumBays()));
    }
    tw.finish();
    ui.update(getExtent());
}

void
WOrderDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size =
        GfxPoint(20 * font_ems[FONT_NORMAL],
                 5 * font_heights[FONT_NORMAL]);
}

void
WOrderDisplay::setBuildOrder(const TBuildOrder& o)
{
    this->o = o;
    drawWidget();
}

static UIWidget&
addFrame(ObjectHolder& h, UIWidget& w)
{
    UIFrameGroup& fg = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    fg.add(w);
    return fg;
}

/**************************** WShipBuildScreen ***************************/

/** \class WShipBuildScreen
    \brief Ship Build Screen, general version
    
    This class provides a complete ship build screen which allows the
    user to select a ship type. It does not provide buttons to confirm
    a build transaction. A derived class must add these. */

WShipBuildScreen::WShipBuildScreen(string_t name, GPlanet& the_planet)
    : UIWindow(name, UIWindow::BLUE_WINDOW, COLOR_DARK),
      hull_list(3000, 10, 15 * font_ems[FONT_NORMAL], &the_planet),
      engine_list(3001, 12 * font_ems[FONT_NORMAL], &the_planet),
      beam_list(3002, 12 * font_ems[FONT_NORMAL], &the_planet),
      torp_list(3003, 12 * font_ems[FONT_NORMAL], &the_planet),
      transfer(the_planet),
      action(the_planet, transfer),
      item_cost_dpy(action, false),
      cards(),
      tabs(cards)
{
    setLayout(UIVBoxLayout::instance5);

    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Horiz)));

    tabs.addTab(1000, UIKeyString(_("Starship Hulls")));
    tabs.addTab(1001, UIKeyString(_("Engines")));
    tabs.addTab(1002, UIKeyString(_("Beams")));
    tabs.addTab(1003, UIKeyString(_("Torpedoes")));

    /* Hulls */
    UIGroup& hull_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1000));
    if (&the_planet) {
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
            if (int j = getTruehull(the_planet.getOwner(), i))
                hull_list.addHullId(j);
    } else {
        for (int i = 1; i <= getNumHulls(); ++i)
            hull_list.addHullId(i);
    }
    hull_group.add(addFrame(h, hull_list));
    hull_group.add(hull_info);
    hull_list.scrollToHull(action.getBuildOrder().hull);
    hull_list.sig_moved.add(this, &WShipBuildScreen::updateHullPage);

    /* Engines */
    UIGroup& engine_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1001));
    engine_group.add(addFrame(h, engine_list));
    engine_group.add(engine_info);
    engine_list.scrollTo(action.getBuildOrder().engine - 1, 1);
    engine_list.sig_moved.add(this, &WShipBuildScreen::updateEnginePage);

    /* Beams */
    UIGroup& beam_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1002));
    beam_group.add(addFrame(h, beam_list));
    beam_group.add(beam_info);
    beam_list.scrollTo(action.getBuildOrder().beam_type - 1, 1);
    beam_list.sig_moved.add(this, &WShipBuildScreen::updateBeamPage);

    /* Torps */
    UIGroup& torp_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1003));
    torp_group.add(addFrame(h, torp_list));
    torp_group.add(torp_info);
    torp_list.scrollTo(action.getBuildOrder().torp_type - 1, 1);
    torp_list.sig_moved.add(this, &WShipBuildScreen::updateLauncherPage);

    cards.add(hull_group);
    cards.add(engine_group);
    cards.add(beam_group);
    cards.add(torp_group);
    cards.sig_focus_changed.add(this, &WShipBuildScreen::updateBuildOrder);

    add(tabs);
    add(cards);

    /* Bill */
    UIGroup& bill_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    bill_group.add(item_cost_dpy);
    bill_group.add(order_dpy);
    add(bill_group);

    hull_list.setFocus();
    updateHullPage();
}

WShipBuildScreen::~WShipBuildScreen()
{
    if (!action.isFinished())
        action.doCancel();
    if (!transfer.isFinished())
        transfer.doCancel();
}

void
WShipBuildScreen::updateHullPage()
{
    hull_info.setHullId(hull_list.getCurrentHull());
    updateBuildOrder();
}

void
WShipBuildScreen::updateBeamPage()
{
    beam_info.setBeamId(beam_list.getCurrent()+1);
    updateBuildOrder();
}

void
WShipBuildScreen::updateLauncherPage()
{
    torp_info.setTorpId(torp_list.getCurrent()+1);
    updateBuildOrder();
}

void
WShipBuildScreen::updateEnginePage()
{
    engine_info.setEngineId(engine_list.getCurrent()+1);
    updateBuildOrder();
}

void
WShipBuildScreen::updateBuildOrder()
{
    TBuildOrder bo;

    if (action.isFinished())  // !
        return;

    bo.hull   = hull_list.getCurrentHull();
    bo.engine = engine_list.getCurrent()+1;
    bo.beam_type  = beam_list.getCurrent()+1;
    bo.beam_count = getHull(bo.hull).getMaxBeams();// !!!!!!!!!!!
    bo.torp_type  = torp_list.getCurrent()+1;
    bo.torp_count = getHull(bo.hull).getMaxLaunchers();// !!!!!!!!!!
    bo.unused = 0;
    action.setBuildOrder(bo);
    order_dpy.setBuildOrder(bo);

    // since this might be called during destruction, we cannot use
    // 'tabs.getCurrentTab()'. 'tabs' might already be gone.
    if (UIWidget* p = cards.getFocusedDescendant()) {
        switch (p->getId()) {
         case 1000: {
            item_cost_dpy.setCost(getHull(hull_list.getCurrentHull()).getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Hull),
                                  getHull(hull_list.getCurrentHull()).getTech());
            break;
         }
         case 1001: {
            GEngine& e = getEngine(engine_list.getCurrent()+1);
            item_cost_dpy.setCost(e.getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Engine), e.getTech());
            break;
         }
         case 1002: {
            GBeam& b = getBeam(beam_list.getCurrent()+1);
            item_cost_dpy.setCost(b.getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Beam), b.getTech());
            break;
         }
         case 1003: {
            GTorpLauncher& t = getTorp(torp_list.getCurrent()+1);
            item_cost_dpy.setCost(t.getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Torpedo), t.getTech());
            break;
         }
        }
    }
}

/*********************** WShipBuildScreenWithActor ***********************/

WShipBuildScreenWithActor::WShipBuildScreenWithActor(GPlanet& the_planet)
    : WShipBuildScreen(_("Build a Ship"), the_planet)
{
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& ok = h.add(new UIButton(2000, _("Enter - Build"), SDLK_RETURN, UIButton::aNothing));
    UIButton& cancel = h.add(new UIButton(2001, _("ESC - Exit"), SDLK_ESCAPE, UIButton::aNothing));
    g.add(ok);
    g.add(cancel);
    if (the_planet.getBaseBuildHull()) {
        UIButton& cancel_build = h.add(new UIButton(2002, UIKeyString(_("C - Cancel Build")), UIButton::aNothing));
        cancel_build.sig_fired.add(this, &WShipBuildScreenWithActor::onCancelBuild);
        g.add(cancel_build);
    }
    g.add(h.add(new UISpacer()));
    add(g);

    ok.sig_fired.add(this, &WShipBuildScreenWithActor::onOk);
    cancel.sig_fired.add(this, &WShipBuildScreenWithActor::onCancel);

    pack();
}

WShipBuildScreenWithActor::~WShipBuildScreenWithActor()
{ }

void
WShipBuildScreenWithActor::onOk()
{
    if (GError e = action.getError()) {
        messageBox(format(_("You can not build this ship: %s"), e.getUserError()),
                   _("Build order rejected"));
        return;
    }
    action.doCommit();
    transfer.doCommit();
    stop(1);
}

void
WShipBuildScreenWithActor::onCancel()
{
    action.doCancel();
    transfer.doCancel();
    stop(0);
}

void
WShipBuildScreenWithActor::onCancelBuild()
{
    static const TBuildOrder blank = { 0, 0, 0, 0, 0, 0, 0 };
    action.doCancel();
    transfer.doCancel();
    action.getPlanet().setBaseBuildOrder(blank);
    stop(1);
}
