/**
  *  \file game/act-base.cc
  *  \brief Building and Scraping Starbases
  *
  *  This contains the GPlanetBuildStarbaseAction class.
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/action.h"
#include "cpluslib/assert.h"
#include "util/nls.h"

/*********************** GPlanetBuildStarbaseAction **********************/

/** \class GPlanetBuildStarbaseAction
    \brief Building and Scraping Starbases

    This class encapsulates the function to build a starbase, or to
    cancel such an order. */

/** Constructor.
    \param the_planet    where we want to build
    \param container     pays that bill
    \param want_build    whether we're building or scraping a base.

    \pre We are playing the planet and the container */
GPlanetBuildStarbaseAction::GPlanetBuildStarbaseAction(GPlanet& the_planet,
                                                       GCargoContainer& container,
                                                       bool want_build)
    : GCargoCostTransaction(container),
      the_planet(the_planet),
      want_build(want_build),
      arb(the_planet.getCargoHandler(), GPlanet::pf_BuildBase)
{
    if (want_build == the_planet.isBuildingBase()) {
        // null operation
        if (want_build)
            throw GError(GError::eDone, _("Starbase building already in progress"));
        else
            throw GError(GError::eDone, _("No starbase being built"));
    } else if (the_planet.hasBase()) {
        // base present
        throw GError(GError::ePerm, _("Starbase already present"));
    } else if (!arb) {
        // concurrency
        throw GError(GError::eFacility, _("This facility is already in use"));
    } else {
        // okay.
        // FIXME: magic numbers
        const int factor = want_build ? +1 : -1;
        setCost(el_Tritanium,  factor * 402);
        setCost(el_Duranium,   factor * 120);
        setCost(el_Molybdenum, factor * 340);
        setCost(el_Money,      factor * 900);
    }
}

/** Destructor. */
GPlanetBuildStarbaseAction::~GPlanetBuildStarbaseAction()
{ }

void
GPlanetBuildStarbaseAction::commit()
{
    ASSERT(isValidCost());
    the_planet.setBuildBaseFlag(want_build);
    GCargoCostTransaction::commit();
}
