/**
  *  \file game/act-sb.cc
  *  \brief General Building on Starbases
  *
  *  This module contains the basic Building on Starbases class, as
  *  well as a number of helper classes.
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/action.h"
#include "game/spec.h"
#include "game/planet.h"
#include "util/nls.h"

/** Cost for upgrading from Tech 1. Indexed by tech level, i.e.
    TECH_COSTS[4] = cost for upgrading from 1 to 4. */
static const int TECH_COSTS[] = {
    0, 0, 100, 300, 600, 1000, 1500, 2100, 2800, 3600, 4500
};

/** \class GStarbaseBuildAction::Actor
    \brief Strategy class for performing orders

    This is a placeholder class for performing the actions of a
    GStarbaseBuildAction. GStarbaseBuildAction descendants provide a
    GStarbaseBuildAction::doChanges() method which performs the
    action we want to do in terms of method calls on an Actor. The
    Actor can now account the action, perform it, or do something
    completely different.

    Note that calls to doChanges() should never call a method twice
    with the same first parameter (e.g. same hull, same tech area,
    etc.). */

GStarbaseBuildAction::Actor::~Actor()
{ }

/** Account a hull which we cannot build. This is called to account
    for hulls which the player cannot build, but which should be
    counted anyway for price computation. */
void
GStarbaseBuildAction::Actor::accountHull(int /*number*/, int /*count*/)
{ }

/***************************** CountingActor *****************************/

/** Counting actor. This actor will just count the costs of all our
    orders, by calling GStarbaseBuildAction::addItemCost() etc. */
class GStarbaseBuildAction::CountingActor : public GStarbaseBuildAction::Actor {
    GStarbaseBuildAction& sba;
 public:
    CountingActor(GStarbaseBuildAction& sba) throw();
    void setTech(TTechLevel area, int value);
    void setHullSlot(int index, int value);
    void setEngine(int index, int value);
    void setBeam(int index, int value);
    void setLauncher(int index, int value);
    void accountHull(int number, int count);
};

inline
GStarbaseBuildAction::CountingActor::CountingActor(GStarbaseBuildAction& sba) throw()
    : sba(sba)
{ }

void
GStarbaseBuildAction::CountingActor::setTech(TTechLevel area, int value)
{
    sba.changeCost(el_Money, TECH_COSTS[value] - TECH_COSTS[sba.the_planet.getBaseTech(area)]);
}

void
GStarbaseBuildAction::CountingActor::setHullSlot(int index, int value)
{
    ASSERT(index > 0 && index <= NUM_HULLS_PER_PLAYER);

    int hull_id = getTruehull(sba.the_planet.getOwner(), index);
    int diff = value - sba.the_planet.getBaseHullStoreSlot(index);

    if (diff) {
        // do this assertion here so users can call setHullSlot for
        // non-existing slots as long as they don't try to change
        // something.
        ASSERT(hull_id != 0);
        sba.addItemCost(::getHull(hull_id).getCost(), diff);
    }
}

void
GStarbaseBuildAction::CountingActor::accountHull(int number, int count)
{
    sba.addItemCost(::getHull(number).getCost(), count);
}

void
GStarbaseBuildAction::CountingActor::setEngine(int index, int value)
{
    ASSERT(index > 0 && index <= NUM_ENGINES);

    sba.addItemCost(::getEngine(index).getCost(),
                    value - sba.the_planet.getBaseEngineStore(index));
}

void
GStarbaseBuildAction::CountingActor::setBeam(int index, int value)
{
    ASSERT(index > 0 && index <= NUM_BEAMS);
    sba.addItemCost(::getBeam(index).getCost(),
                    value - sba.the_planet.getBaseBeamStore(index));
}

void
GStarbaseBuildAction::CountingActor::setLauncher(int index, int value)
{
    ASSERT(index > 0 && index <= NUM_TORPS);
    sba.addItemCost(::getTorp(index).getCost(),
                    value - sba.the_planet.getBaseLauncherStore(index));
}

/***************************** ExecutingActor ****************************/

/** Executing actor. This actor will write all changes to the
    planet we're working on. */
class GStarbaseBuildAction::ExecutingActor : public GStarbaseBuildAction::Actor {
    GStarbaseBuildAction& sba;
 public:
    ExecutingActor(GStarbaseBuildAction& sba) throw();
    void setTech(TTechLevel area, int value);
    void setHullSlot(int index, int value);
    void setEngine(int index, int value);
    void setBeam(int index, int value);
    void setLauncher(int index, int value);
};

inline
GStarbaseBuildAction::ExecutingActor::ExecutingActor(GStarbaseBuildAction& sba) throw()
    : sba(sba)
{ }

void
GStarbaseBuildAction::ExecutingActor::setTech(TTechLevel area, int value)
{
    sba.the_planet.setBaseTech(area, value);
}

void
GStarbaseBuildAction::ExecutingActor::setHullSlot(int index, int value)
{
    sba.the_planet.setBaseHullStoreSlot(index, value);
}

void
GStarbaseBuildAction::ExecutingActor::setEngine(int index, int value)
{
    sba.the_planet.setBaseEngineStore(index, value);
}

void
GStarbaseBuildAction::ExecutingActor::setBeam(int index, int value)
{
    sba.the_planet.setBaseBeamStore(index, value);
}

void
GStarbaseBuildAction::ExecutingActor::setLauncher(int index, int value)
{
    sba.the_planet.setBaseLauncherStore(index, value);
}

/************************** GStarbaseBuildAction *************************/

/** \class GStarbaseBuildAction
    \brief Accounting for Building on Starbases.

    This class provides accounting for all sorts of build orders on a
    starbase. It does not provide rule checks.

    Derived classes will override doChanges(). GStarbaseBuildAction
    will use that method to compute the costs, and to actually execute
    the commands if so desired. You must call regenerate() whenever
    you change something (so that we get a chance to re-compute the
    costs). regenerate() will raise sig_changed. */

/** Constructor.
    \param the_planet  this is the starbase
    \param container   this container will pair our expenses. Usually,
                       this is the planet transporter. When you commit
                       this transaction, you must commit the container
                       as well. */
GStarbaseBuildAction::GStarbaseBuildAction(GPlanet& the_planet, GCargoContainer& container)
    : GCargoCostTransaction(container), the_planet(the_planet),
      arb(the_planet.getCargoHandler(), GPlanet::pf_BaseConstruction)
{
    if (!the_planet.hasBase())
        throw GError(GError::eNoBase, _("The planet does not have a starbase"));
    if (!arb)
        throw GError(GError::eFacility, _("This facility is already in use"));
}

GStarbaseBuildAction::~GStarbaseBuildAction()
{ }

/** Recompute the costs and update our reservations. */
void
GStarbaseBuildAction::regenerate()
{
    clearCosts();
    if (the_planet.hasBase()) {
        CountingActor ca(*this);
        doChanges(ca);
        sig_changed.raise();
    }
}

void
GStarbaseBuildAction::commit()
{
    regenerate();
    ASSERT(!getError());

    ExecutingActor xa(*this);
    doChanges(xa);
    GCargoCostTransaction::commit();
}

void
GStarbaseBuildAction::cancel()
{
    GCargoCostTransaction::cancel();
}

/** Add cost of a number of items.
    \param c    item
    \param diff number of items (can be negative if you're selling). */
void
GStarbaseBuildAction::addItemCost(TCost c, int32 diff)
{
    changeCost(el_Money,      c.mc * diff);
    changeCost(el_Tritanium,  c.t * diff);
    changeCost(el_Molybdenum, c.m * diff);
    changeCost(el_Duranium,   c.d * diff);
}
