/**
  *  \file game/action.h
  *  \brief Game Actions
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ACTION_H
#define PCC_V2_GAME_ACTION_H

#include "game/planet.h"
#include "game/cargo.h"
#include "game/struct.h"

class GObjectList;

/** \defgroup actions Game Actions

    These classes encapsulate game actions. They are used by the GUI
    code as well as the interpreter.

    Only nontrivial actions get their own class. Some actions have
    complicated preconditions, or interact with different objects;
    such things need an action class. Just changing a mission or
    friendly code can be done directly. */
//@{

/* act-base.cc */
class GPlanetBuildStarbaseAction : public GCargoCostTransaction {
    GPlanet&  the_planet;
    bool      want_build;
    GCargo::FacilityArbiter arb;
 public:
    GPlanetBuildStarbaseAction(GPlanet& the_planet, GCargoContainer& container, bool want_build);
    ~GPlanetBuildStarbaseAction();
 protected:
    void commit();
};

/* act-structures.cc */
//! Building structures on a planet.
class GPlanetBuildStructuresAction : public GCargoCostTransaction {
    enum { MAX = str_BaseDefense + 1 };
    struct Data {
        int   order;            ///< Amount selected by used.
        int   max;              ///< Maximum amount allowed by rules.
        int   min;              ///< Minimum amount allowed by rules (undo).
    };
    Data      data[MAX];
    GPlanet&  the_planet;
    GCargo::FacilityArbiter arb;

    void computeCost();

 protected:
    void commit();

 public:
    GPlanetBuildStructuresAction(GPlanet& the_planet, GCargoContainer& container);
    ~GPlanetBuildStructuresAction();

    void setUndoInformation(const GPlanet& dis_planet);

    int addStructures(TPlanetaryStructure ps, int count, bool partial);
    int addStructuresLimitCash(TPlanetaryStructure ps, int count);

    /** Get planet we're working on. */
    GPlanet& getPlanet()
        { return the_planet; }

    /** Get minimum amount for a structure. Structures cannot be destroyed. */
    int getMin(TPlanetaryStructure ps) const
        { return data[ps].min; }
    /** Get maximum amount for a structure. */
    int getMax(TPlanetaryStructure ps) const
        { return data[ps].max; }
    /** Get current ordered amount for a structure. */
    int getAmount(TPlanetaryStructure ps) const
        { return data[ps].order; }
};

/* act-supplies.cc */
class GPlanetSellSuppliesAction : public GTransaction {
    GPlanet&         the_planet;
    GPlanetTransfer  transfer;
    int32            amount;
    int32            reserved_sup;
 public:
    GPlanetSellSuppliesAction(GPlanet& the_planet);
    ~GPlanetSellSuppliesAction();

    int32 getMinimum() const;
    int32 getMaximum() const;

    void  setAmount(int32 amount);
    int32 getAmount() const
        { return amount; }

 protected:
    void commit();
    void cancel();
};

/* act-sb.cc */
class GStarbaseBuildAction : public GCargoCostTransaction {
    GPlanet&          the_planet;
    GCargo::FacilityArbiter arb;
 public:
    class Actor : public Object {
     public:
        virtual ~Actor();
        virtual void setTech(TTechLevel area, int value) = 0;
        virtual void setHullSlot(int index, int value) = 0;
        virtual void setEngine(int index, int value) = 0;
        virtual void setBeam(int index, int value) = 0;
        virtual void setLauncher(int index, int value) = 0;
        virtual void accountHull(int number, int count);
    };

    GStarbaseBuildAction(GPlanet& the_planet, GCargoContainer& container);
    ~GStarbaseBuildAction();

    void regenerate();
    
    GPlanet& getPlanet()
        { return the_planet; }
 protected:
    virtual void doChanges(Actor& actor) = 0;

    void commit();
    void cancel();

 private:
    class CountingActor;
    class ExecutingActor;
    friend class CountingActor;
    friend class ExecutingActor;
    void addItemCost(TCost c, int32 diff);
};

/* act-tech.cc */
class GStarbaseTechUpgradeAction : public GStarbaseBuildAction {
    GPlanet& the_planet;
    int      new_tech[4], min_tech[4];

    void     doChanges(Actor& actor);
 public:
    GStarbaseTechUpgradeAction(GPlanet& the_planet, GCargoContainer& container);
    ~GStarbaseTechUpgradeAction();

    bool     isValid(int index, int new_val) const;
    bool     setTech(int index, int new_val);
    int      getCost() const { return GStarbaseBuildAction::getCost(el_Money); }

    int      getTech(int index) const    { return new_tech[index]; }
    int      getMinTech(int index) const { return min_tech[index]; }
    int      getMaxTech(int index) const;

    int      getMissingCash() const { return getMissing(el_Money); }
    int      getReservedCash() const { return getReserved(el_Money); }
    int      getReservedSupplies() const { return getReserved(el_Supplies); }
};

/* act-fix.o */
class GStarbaseFixRecycleAction {
    int        num_own, num_foreign;
    GUniverse& univ;
    GPlanet&   planet;
 public:
    GStarbaseFixRecycleAction(GUniverse& univ, GPlanet& planet);
    bool canFix() const;
    bool canRecycle() const;
    bool setAction(TShipyardAction action, int ship_id);
    void enumerateShipsFor(TShipyardAction action, GObjectList& list);
};

/* act-ammo.o */
/** Ammunition building. This transaction allows to build ammunition
    from a planet, and place it onto any receiver object. */
class GStarbaseBuildAmmoAction : public GCargoCostTransaction {
 public:
    enum {
        /// Minimum index for accessor functions.
        min_index = 1,
        /// Maximum index for accessor functions.
        max_index = NUM_TORPS+1
    };
 private:
    /// Planet who will pay for new ammo.
    GPlanetTransfer& pay;
    /// Receiver of new ammo.
    GCargoContainer& receiver;
    /// Amount of ammo added.
    int32 added[max_index];

 protected:
    void   commit();
    void   cancel();

 public:
    GStarbaseBuildAmmoAction(GPlanetTransfer& pay, GCargoContainer& receiver);
    ~GStarbaseBuildAmmoAction();

    bool   canHave(int index) const;
    int32  getAmount(int index) const;
    void   setAmount(int index, int32 amount);
    int    getTechUpgradeNeeded(int index) const;

    GPlanet& getPlanet()
        { return pay.getPlanet(); }

    static GCargoType getCargoTypeFromIndex(int index);
};

/* act-build.cc */
class GStarbaseBuildShipAction : public GStarbaseBuildAction {
    TBuildOrder        order;
    bool               use_parts_from_storage : 1;
    bool               use_tech_upgrade       : 1;
    bool               is_ok                  : 1;

    void  doTechUpgrade(TTechLevel tl, Actor& a, int need);
    int   getBuildAmount(int need, int have);

 protected:
    void  commit();

 public:
    GStarbaseBuildShipAction(GPlanet& planet, GPlanetTransfer& pay);
    ~GStarbaseBuildShipAction();

    bool  isUsePartsFromStorage() const { return use_parts_from_storage; }
    bool  isUseTechUpgrade() const { return use_tech_upgrade; }

    void  setUsePartsFromStorage(bool b);
    void  setUseTechUpgrade(bool b);

    void                setBuildOrder(const TBuildOrder& o);
    const TBuildOrder&  getBuildOrder() const { return order; }

    void  doChanges(Actor& a);

    GError getError();
};

//@}

#endif
