/**
  *  \file game/coord.cc
  *  \brief Game Coordinate System Handling
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <iostream>
#include "game/coord.h"
#include "game/config.h"
#include "game/host.h"

/** Global starchart configuration. */
GChartConfiguration chart_conf = {
    false,
    1000, 1000,                 // min
    3000, 3000,                 // max
    2000, 2000                  // size
};

/********************************* GPoint ********************************/

/** Check if point is on map.
    Consistent with behaviour of PCC 1.x, PHost 3.3c, and EchoView. */
bool
GPoint::isOnMap() const
{
    if (!chart_conf.wrapped) {
        return true;
    } else {
        return (x < chart_conf.xmax && x >= chart_conf.xmin
                && y < chart_conf.ymax && y >= chart_conf.ymin);
    }
}

/** Get normalized version of this point.
    The incarnation of this point that is on the main tile of the map. */
GPoint
GPoint::getNormalized() const
{
    int x = this->x, y = this->y;
    if (chart_conf.wrapped) {
        if (x >= chart_conf.xmax)
            x -= chart_conf.xsize;
        if (x <  chart_conf.xmin)
            x += chart_conf.xsize;
        if (y >= chart_conf.ymax)
            y -= chart_conf.ysize;
        if (y <  chart_conf.ymin)
            y += chart_conf.ysize;
    }
    return GPoint(x, y);
}

/** Get incarnation nearest to a. On a wrapped map, returns
    the incarnation of this point that is nearest to a.
    Usage: player clicks into map at point a, and we find that
    planet X is near to it. The equivalent coordinates are
    X.getPos().getNearestTo(a), which accounts for the fact
    that the player may have clicked at the "wrong" side of the map
    border. */
GPoint
GPoint::getNearestTo(GPoint a) const
{
    int x = this->x, y = this->y;
    if (chart_conf.wrapped) {
        if (2*(a.x - x) > chart_conf.xsize)
            x += chart_conf.xsize;
        if (2*(a.y - y) > chart_conf.ysize)
            y += chart_conf.ysize;
        if (2*(x - a.x) > chart_conf.xsize)
            x -= chart_conf.xsize;
        if (2*(y - a.y) > chart_conf.ysize)
            y -= chart_conf.ysize;
    }
    return GPoint(x, y);
}

/** Get sector number this point is in, as decimal number. This
    returns the same sector number as in PCC 1.x, which happens to be
    the same as the sectors on Trevor Fuson's VGAMAP for a
    standard-sized map.
    \return sector number (100..499), zero if point is not in any
    numbered sector */
int
GPoint::getSectorNumber() const
{
    int x = this->x - (chart_conf.xmax - chart_conf.xmin) / 2 + 1000;
    int y = this->y - (chart_conf.ymax - chart_conf.ymin) / 2 + 1000;

    /* outside known region? */
    if (x < 0 || y < 0 || x >= 2000 || y >= 2000)
        return 0;

    int major = 1;
    if (x >= 1000)
        major += 1;
    if (y < 1000)
        major += 2;

    return 100 * major
        + 10 * ((x % 1000) / 100)
        + (999 - (y % 1000)) / 100;
}

/************************** GChartConfiguration **************************/

/** Initialize configuration. Validates the configuration and computes
    the derived information. */
void
GChartConfiguration::init()
{
    if (xmax < xmin)
        std::swap(xmax, xmin);
    if (ymax < xmin)
        std::swap(ymax, ymin);

    xsize = xmax - xmin;
    ysize = ymax - ymin;
}

/******************************* Utilities *******************************/

/** Get minimum distance between two points, squared. */
long
distanceSquared(GPoint a, GPoint b)
{
    GPoint c = b.getNearestTo(a);
    return distanceSquaredRaw(a, c);
}

/** Get distance between two points, squared. */
long
distanceSquaredRaw(GPoint a, GPoint b)
{
    long x = a.x - b.x;
    long y = a.y - b.y;
    return x*x + y*y;
}

/** Initialize. Assumes configuration has been read. */
void
initCoordinates()
{
    if (host.isPHost()) {
        chart_conf.wrapped = config.AllowWraparoundMap;
        chart_conf.xmin = config.WraparoundRectangle[0];
        chart_conf.ymin = config.WraparoundRectangle[1];
        chart_conf.xmax = config.WraparoundRectangle[2];
        chart_conf.ymax = config.WraparoundRectangle[3];
    }
    chart_conf.init();
}

/** Output GPoint to a stream. */
std::ostream&
operator<<(std::ostream& os, GPoint pt)
{
    return os << '(' << pt.x << ", " << pt.y << ')';
}
