/**
  *  \file game/error.h
  *  \brief Error Messages
  *
  *  This file declares the GError class which contains an error
  *  message.
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ERROR_H
#define PCC_V2_GAME_ERROR_H

#include <exception>

/** Error message. This class provides a way to store error messages
    in a format usable by both the script and user side. */
class GError : public std::exception {
    const char* script_error;
    const char* user_error;
 public:
    /** Constructor. Create a "no-error" status. */
    GError()
        : script_error(0), user_error(0)
        { }
    /** Constructor. Create an error message.
        \param script_error  Script error message. Ideally, one of the eXxxx constants.
        \param user_error    User error message. Should be a translated string. */
    GError(const char* script_error, const char* user_error)
        : script_error(script_error), user_error(user_error)
        { }

    /** Get script-side error message. */
    const char* getScriptError() const
        { return script_error; }
    /** Get user error message. */
    const char* getUserError() const
        { return user_error; }
    /** Get user error message. */
    const char* what() const throw()
        { return user_error; }

    /** Convert to bool.
        \return true if there was an error. */
    operator bool() const
        { return script_error != 0; }

    bool isError(const char* which_one) const;

    static const char eFacility[], eNotOwner[], eRange[],
        ePerm[], eNoBase[], eNoResource[], ePos[], ePartial[],
        eUser[], eFleet[], eGraph[], eDone[], eNotPlaying[];
};

#endif
