/**
  *  \file game/global-bb.cc
  *  \brief Chart Bounding Boxes
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/ion.h"
#include "game/mine.h"

/******************************* GChartBBox ******************************/

/*! \class GChartBBox
    \brief Computing a Starchart Bounding Box

    This class provides functions to compute a bounding box. The sequence of
    events is:
    - add points using addPoint(), addCircle()
    - call initMaxZoom(), initMinZoom() to make the final computations */

/** Initialize bounding box from game universe. */
void
GChartBBox::initFromUniverse(GUniverse& univ)
{
    min = GPoint(chart_conf.xmin, chart_conf.ymin);
    max = GPoint(chart_conf.xmax, chart_conf.ymax);

    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const GPlanet& p = univ.getPlanet(i);
        if (p.isExistant())
            addPoint(p.getPos());
    }
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& s = univ.getShip(i);
        if (s.isExistant())
            addPoint(s.getPos());
    }
    for (int i = 1; i <= NUM_STORMS; ++i) {
        const GIonStorm& st = univ.getIonStorm(i);
        if (st.isExistant())
            addCircle(st.getPos(), st.getRadius());
    }
    for (GUniverse::MinefieldMap::const_iterator i = univ.getMinefieldBegin(); i != univ.getMinefieldEnd(); ++i)
        addCircle(i->second->getPos(), i->second->getRadius());

    initMinZoom();
    initMaxZoom();
}

/** Compute minimum zoom level. The minimum zoom is a soft limit;
    below the limit, the scrolling map is replaced by a fixed one. */
void
GChartBBox::initMinZoom()
{
    /* Minimum zoom. At zoom smaller 1:N, switch to overview chart */
    min_zoom = 2;
    // FIXME: dummy
    // CheckMinZoom(SCREEN_WIDTH DIV 2, ChartBX1-ChartConf.CenterX);
    // CheckMinZoom(SCREEN_WIDTH DIV 2, ChartConf.CenterX-ChartBX0);
    // CheckMinZoom((SCREEN_HEIGHT-80) DIV 2, ChartBY1-ChartConf.CenterY);
    // CheckMinZoom((SCREEN_HEIGHT-80) DIV 2, ChartConf.CenterY-ChartBY0);
    if (min_zoom > MAX_ZOOM)
        min_zoom = MAX_ZOOM;
}

/** Compute maximum zoom level. The maximum zoom level is a hard
    limit; under no conditions the hard limit may be exceeded to avoid
    integer overflow. */
void
GChartBBox::initMaxZoom()
{
    max_zoom = INT_MAX / std::max(max.x, max.y);
    if (max_zoom > MAX_ZOOM)
        max_zoom = MAX_ZOOM;
}

/** Include a point in the bounding box. */
void
GChartBBox::addPoint(GPoint pt)
{
    if (min.x > pt.x)
        min.x = pt.x;
    if (min.y > pt.y)
        min.y = pt.y;
    if (max.x < pt.x)
        max.x = pt.x;
    if (max.y < pt.y)
        max.y = pt.y;
}

/** Include a circle in the bounding box. */
void
GChartBBox::addCircle(GPoint pt, int radius)
{
    addPoint(GPoint(pt.x - radius, pt.y - radius));
    addPoint(GPoint(pt.x + radius, pt.y + radius));
}
