/**
  *  \file game/objtypes.cc
  *  \brief Object types
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objtypes.h"
#include "game/obj.h"
#include "cpluslib/assert.h"

/****************************** GObjectType ******************************/

/** \class GObjectType
    \brief Object Type

    A class derived from GObjectType defines a set of objects, and
    iteration through them. In addition, it contains a general
    interface for object selection. Control screens and various other
    features use these as a back-end.

    A GObjectType object also manages a "current position".

    \todo We need to specify semantics for object Ids. The most
    sensible interpretation would be to have no forced interpretation,
    and let GPlanetType etc. guarantee the additional requirement that
    id == got.getObject(id).getId(). Not having this requirement
    simplifies implementing the list for the visual scanner, for
    example. */

/** Initialize.
    \param max  Maximum object Id.
    \param trn  Related universe object. */
GObjectType::GObjectType(int max, GUniverse& univ)
    : univ(univ), current(0), count(0), max(max)
{ }

GObjectType::~GObjectType()
{ }

/** Initialize. Call this after any change to the object set. Derived
    classes can override this method to provide additional
    initialisation. */
void
GObjectType::init()
{
    count = 0;
    if (!isExistant(current))
        current = 0;
    for (int i = 1; i <= max; ++i) {
        if (isExistant(i)) {
            ++count;
            if (!current)
                current = i;
        }
    }
}

/** Get next object after /index/.
    \param index current object number
    \param mark iff true, only access marked objects
    \returns 0 if no object (only if count=0 or mark requested and no object marked!)
    \default goes through objects in Id order. */
int
GObjectType::getNext(int index, bool mark)
{
    bool wrapped = false;
    while (1) {
        if (++index > max) {
            index = 0;
            if (wrapped)
                return 0;
            wrapped = 1;
        }
        if (isExistant(index) && (!mark || getObject(index).isMarked()))
            return index;
    }
}

/** Get previous object before /index/.
    \param index current object number
    \param mark iff true, only access marked objects
    \returns 0 if no object (only if count=0 or mark requested and no object marked!)
    \default goes through objects in Id order. */
int
GObjectType::getPrev(int index, bool mark)
{
    bool wrapped = false;
    while (1) {
        if (--index <= 0) {
            index = max;
            if (wrapped)
                return 0;
            wrapped = 1;
        }
        if (isExistant(index) && (!mark || getObject(index).isMarked()))
            return index;
    }
}

/** Set current object Id.
    \pre isExistant(index) */
void
GObjectType::setCurrent(int index)
{
    ASSERT(isExistant(index));
    if (current != index) {
        current = index;
        sig_pos_changed.raise(index);
    }
}
