/**
  *  \file game/un-trnflt.h
  *  \brief Filter Classes for Un-TRN
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UN_TRNFLT_H
#define PCC_V2_GAME_UN_TRNFLT_H

#include "game/trn.h"
#include "util/except.h"
#include "cpluslib/stringcv.h"
#include "cpluslib/misc.h"
#include "cpluslib/holder.h"

//! Syntax error while parsing a filter expression.
class ParseException : public std::runtime_error {
 public:
    ParseException(const string_t& what)
        : std::runtime_error(toStandardString(what)) { }
};

/** \defgroup Filter Turn Command Filters

    un-trn filter expressions are represented by trees of Filter
    descendants. This group contains the classes as well as functions
    to parse them.

    Memory management is done by ObjectHolder. */
//@{

//! Abstract Filter Class.
class Filter : public Object {
 public:
    /** Check whether command is accepted.
        \param trn turn file we're looking at
        \param index command number we're looking at, 0-based */
    virtual bool accept(GTurnfile& trn, int index) = 0;
};

//! Constant result (permanently true or false).
class FilterConst : public Filter {
    bool res;
 public:
    /** Create.
        \param res result to return */
    FilterConst(bool res) : res(res) { }
    virtual bool accept(GTurnfile& trn, int index);
};

//! Filtering for command names.
class FilterName : public Filter {
    bool wild;
    string_t str;
 public:
    /** Create.
        \param str name we want to match
        \param wild true to match all commands beginning with that
                    name, false to match exact names */
    FilterName(string_t str, bool wild = false)
        : wild(wild), str(str) { }
    bool accept(GTurnfile& trn, int index);
};

//! Filtering for string contents.
class FilterString : public Filter {
    string_t str;
 public:
    /** Create.
        \param str string to match (case-insensitive) */
    FilterString(string_t str)
        : str(strUCase(str)) { }
    bool accept(GTurnfile& trn, int index);
};

//! Filtering for object Ids.
class FilterId : public Filter {
    int lower, upper;
 public:
    /** Create.
        \param low,up range to match, boundaries inclusive. */
    FilterId(int low, int up)
        : lower(low), upper(up) { }
    bool accept(GTurnfile& trn, int index);
};

//! Conjunction of two filter expressions.
class FilterAnd : public Filter {
    Filter& lhs;
    Filter& rhs;
 public:
    /** Create.
        \param lhs,rhs sub-expressions. Both must be
        true that we return true. */
    FilterAnd(Filter& lhs, Filter& rhs)
        : lhs(lhs), rhs(rhs) { }
    bool accept(GTurnfile& trn, int index);
};

//! Disjunction of two filter expressions.
class FilterOr : public Filter {
    Filter& lhs;
    Filter& rhs;
 public:
    /** Create.
        \param lhs,rhs sub-expressions. Both must be
        false that we return false. */
    FilterOr(Filter& lhs, Filter& rhs)
        : lhs(lhs), rhs(rhs) { }
    bool accept(GTurnfile& trn, int index);
};

//! Negation of a filter expression.
class FilterNeg : public Filter {
    Filter& other;
 public:
    /** Create.
        \param o sub-expression. We return true iff it
        returns false. */
    FilterNeg(Filter& o)
        : other(o) { }
    bool accept(GTurnfile& trn, int index);
};

//! Match command indexes.
class FilterIndex : public Filter {
    int from, to;
 public:
    /** Create.
        \param from,to index range, 1-based, boundaries inclusive. */
    FilterIndex(int from, int to)
        : from(from), to(to) { }
    bool accept(GTurnfile& trn, int index);
};

Filter& parseFilter(const char* text, ObjectHolder& h);
//@}

#endif
