/**
  *  \file util/cmdline.h
  *  \brief Command Line Handling
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_CMDLINE_H
#define PCC_V2_UTIL_CMDLINE_H

#include <vector>
#include <cstddef>
#include "util/except.h"

//! Command line parser.
class CommandLineParser {
    char** argv;
    static const char* prog;

    const char* arg;
    bool traditional;

    const char* getNext();
 public:
    CommandLineParser(bool trad = false);
    virtual ~CommandLineParser();
    void parse(char** aargv);

    virtual void parseStarted();
    virtual void parseFinished();
    virtual void handleNonoption(const string_t&) = 0;
    virtual void handleOption(const string_t&, bool have_arg) = 0;

    const char* getArg();

    /** Get program name. This is filled in when parse() is first
        called. */
    static const char* getProgramName()
        { return prog; }
};

class StandardCommandLineParser : public CommandLineParser {
 public:
    /** Definition of an option. */
    struct Option {
        /** Name. */
        const char* name;
        /** Function to be called when option is encountered. */
        void        (*handler)(StandardCommandLineParser& parser, const Option& opt);
        /** User-defined value. Not used by StandardCommandLineParser, can be
            used by the function. */
        void*       user_data;
        /** Whether the option takes a parameter. */
        bool        takes_arg;
        /** Help text, untranslated. */
        const char* help;
    };
    StandardCommandLineParser(bool trad = false);
    ~StandardCommandLineParser();
    void addOptionSet(const Option* opt, std::size_t count);

    void handleOption(const string_t& opt, bool have_arg);
    void printHelp(std::ostream& os);

    virtual void handleError(const string_t& msg);
 private:
    std::vector<Option> options;
};

/** Command line error. CommandLineParser throws this if it sees an
    invalid option. */
class CommandLineException : public UserException {
 public:
    CommandLineException(const string_t& str)
        : UserException(str) { }
};

/* Some default handlers */

/** Command line option handler: Store fixed value. opt.user_data must point
    at the variable.
    \param T     type of variable
    \param value value to store */
template <class T, T value>
void optSetValue(StandardCommandLineParser&, const StandardCommandLineParser::Option& opt)
{
    *static_cast<T*>(opt.user_data) = value;
}

/** Command line option handler: Call a free function.
    \param func function to call. */
template <void (*func)()>
void optCallFunction(StandardCommandLineParser&, const StandardCommandLineParser::Option&)
{
    func();
}

#endif
